# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Synchronization of local datasets with drafst on Zenodo.

This script synchronizes your local version of a dataset
with the (draft of) this dataset on Zenodo.
This simplifies your interaction with Zenodo to the following steps:

1. Prepare a personal token
2. Write metadata in a configuration YAML file.
3. Include this script in your workflow to automatically sync data.
4. Publish the dataset through the Zenodo web interface.

After publication, only metadata can be updated, not the actual files.
If you need to upload different files, increment the version number.
This will result in a new draft that you can publish when it is final.

This script offers a simplified interface to Zenodo.
It does not try to mirror all of Zenodo's API features.

Note that the current version uses the Invenio RDM API (v12) to interact with Zenodo.
This API is not officially documented by Zenodo, but it is already used by Zenodo itself.
Relevant API documentation can be found here:

- https://inveniordm.docs.cern.ch/reference/rest_api_index/

Unofficial information on how to use the Invenio RDM API with Zenodo:

- https://github.com/zenodo/developers.zenodo.org/issues/62
- https://github.com/zenodo/zenodo/issues/2544

Note that the new Zenodo API (based on Invenio RDM) is still not fully stable.
Some of the features implemented in this module were deduced from the Zenodo web interface.
"""

import argparse
import datetime
import hashlib
import json
import sys
from typing import Any

import attrs
import cattrs
import idutils
import requests
import semver
import yaml
from markdown_it import MarkdownIt
from path import Path
from rich import print  # noqa: A004

from stepup.core.api import amend, getenv


class RESTError(Exception):
    """Raised when a REST API call is not successful."""


class ZenodoError(Exception):
    """Raised in case of an error in the logic of the main script."""


@attrs.define
class RESTWrapper:
    """Wrapper for REST APIs that use JSON input (mostly) and output (always)."""

    endpoint: str = attrs.field()
    params: dict[str, str] = attrs.field()
    verbose: bool = attrs.field(default=False)

    def request(self, method: str, loc: str, **kwargs) -> Any:
        """Send a HTTP request and deserialize the response as JSON.

        Parameters
        ----------
        method
            The HTTP method: GET, POST, PUT or DELETE.
        loc
            The address to be appended after the endpoint.
        kwargs
            Keyword arguments to pass on to the `requests.request` function.

        Returns
        -------
        response_data
            Deserialized JSON response data.
        """
        url = f"{self.endpoint}/{loc}"
        if self.verbose:
            print(f"{method} {url}")
            if "json" in kwargs:
                print("REQUEST")
                print(kwargs["json"])
        res = requests.request(method, url, params=self.params, **kwargs)
        data = None if len(res.text) == 0 else res.json()
        if not res.ok:
            raise RESTError(
                f"Failed {method} {url}: {res.status_code}\n" + json.dumps(data, indent=2)
            )
        if self.verbose:
            print("RESPONSE")
            print(data)
            print()
        return data

    def get(self, loc: str, **kwargs):
        """Create a GET HTTP requests. See `request` method for details."""
        return self.request("GET", loc, **kwargs)

    def post(self, loc: str, **kwargs):
        """Create a POST HTTP requests. See `request` method for details."""
        return self.request("POST", loc, **kwargs)

    def put(self, loc: str, **kwargs):
        """Create a PUT HTTP requests. See `request` method for details."""
        return self.request("PUT", loc, **kwargs)

    def delete(self, loc: str, **kwargs):
        """Create a DELETE HTTP requests. See `request` method for details."""
        return self.request("DELETE", loc, **kwargs)


# See https://datacite-metadata-schema.readthedocs.io/en/4.6/properties/resourcetype/
RESOURCE_TYPES = [
    "audio",
    "dataset",
    "event",
    "image",
    "image-diagram",
    "image-drawing",
    "image-figure",
    "image-other",
    "image-photo",
    "image-plot",
    "lesson",
    "model",
    "other",
    "physicalobject",
    "poster",
    "presentation",
    "publication",
    "publication-annotationcollection",
    "publication-article",
    "publication-book",
    "publication-conferencepaper",
    "publication-conferenceproceeding",
    "publication-datamanagementplan",
    "publication-datapaper",
    "publication-deliverable",
    "publication-dissertation",
    "publication-journal",
    "publication-milestone",
    "publication-other",
    "publication-patent",
    "publication-peerreview",
    "publication-preprint",
    "publication-proposal",
    "publication-report",
    "publication-section",
    "publication-softwaredocumentation",
    "publication-standard",
    "publication-taxonomictreatment",
    "publication-technicalnote",
    "publication-thesis",
    "publication-workingpaper",
    "software",
    "software-computationalnotebook",
    "video",
    "workflow",
]

# See https://github.com/inveniosoftware/invenio-rdm-records/blob/master/invenio_rdm_records/config.py
IDENTIFIER_SCHEMES = {
    "ark": idutils.is_ark,
    "arxiv": idutils.is_arxiv,
    "ads": idutils.is_ads,
    "crossreffunderid": lambda _: True,
    "doi": idutils.is_doi,
    "ean13": idutils.is_ean13,
    "eissn": idutils.is_issn,
    "grid": lambda _: True,
    "handle": idutils.is_handle,
    "igsn": lambda _: True,
    "isbn": idutils.is_isbn,
    "isni": idutils.is_isni,
    "issn": idutils.is_issn,
    "istc": idutils.is_istc,
    "lissn": idutils.is_issn,
    "lsid": idutils.is_lsid,
    "pmid": idutils.is_pmid,
    "purl": idutils.is_purl,
    "upc": lambda _: True,
    "url": idutils.is_url,
    "urn": idutils.is_urn,
    "w3id": lambda _: True,
    "other": lambda _: True,
}

# See https://github.com/inveniosoftware/invenio-rdm-records/blob/master/invenio_rdm_records/fixtures/data/vocabularies/relation_types.yaml
RELATION_TYPES = [
    "cites",
    "compiles",
    "continues",
    "describes",
    "documents",
    "hasmetadata",
    "haspart",
    "hasversion",
    "iscitedby",
    "iscompiledby",
    "iscontinuedby",
    "isderivedfrom",
    "isdescribedby",
    "isdocumentedby",
    "isidenticalto",
    "ismetadatafor",
    "isnewversionof",
    "isobsoletedby",
    "isoriginalformof",
    "ispartof",
    "ispreviousversionof",
    "ispublishedin",
    "isreferencedby",
    "isrequiredby",
    "isreviewedby",
    "issourceof",
    "issupplementedby",
    "issupplementto",
    "isvariantformof",
    "isversionof",
    "obsoletes",
    "references",
    "requires",
    "reviews",
]


@attrs.define
class Organization:
    """A subset of Invenio RDM affiliation / funder."""

    name: str | None = attrs.field(default=None)
    """Name of the affiliation, e.g. a university or research institute."""

    ror: str | None = attrs.field(default=None)
    """The Research Organization Registry identifier."""

    @ror.validator
    def _validate_ror(self, attribute, value):
        """Validate the ROR identifiers."""
        if not (value is None or idutils.is_ror(value)):
            raise ValueError("Invalid ROR identifier format.")

    def __attrs_post_init__(self):
        if not ((self.name is None) ^ (self.ror is None)):
            raise ValueError("Exactly one of 'name' or 'ror' must be set for an affiliation.")

    def to_zenodo(self) -> dict[str, Any]:
        """Convert the affiliation to a dictionary suitable for Zenodo."""
        result = {}
        if self.name is not None:
            result["name"] = self.name
        if self.ror is not None:
            result["id"] = self.ror
        if len(result) != 1:
            raise ValueError("Exactly one of 'name' or 'ror' must be set for an affiliation.")
        return result


@attrs.define
class Creator:
    """A subset of Invenio RDM creator."""

    family_name: str | None = attrs.field(default=None)
    """Family name."""

    given_name: str | None = attrs.field(default=None)
    """Given name."""

    identifiers: dict[str, str] = attrs.field(factory=dict)
    """Identifiers of the creator, e.g. ORCID or ISNI.

    Keys must be lower case strings.
    Values must not contain a https:// prefix, but only the identifier itself.
    For example, for ORCID, use '0000-0001-9288-5608'
    instead of 'https://orcid.org/0000-0001-9288-5608'.
    """

    @identifiers.validator
    def _validate_identifiers(self, attribute, value):
        """Validate the identifiers."""
        if not isinstance(value, dict):
            raise TypeError("Identifiers must be a dictionary.")
        for key in value:
            if key not in ["orcid", "isni"]:
                raise ValueError(
                    f"Unknown identifier type: {key}. Only 'orcid' or 'isni' are allowed."
                )
        if "orcid" in value:
            if not idutils.is_orcid(value["orcid"]):
                raise ValueError(f"Invalid ORCID: {value['orcid']}")
            if value["orcid"].startswith("http"):
                raise ValueError(
                    "ORCID identifiers must not start with 'http'. "
                    "Use only the identifier itself, e.g. '0000-0001-9288-5608'."
                )
        if "isni" in value:
            if not idutils.is_isni(value["isni"]):
                raise ValueError(f"Invalid ISNI: {value['isni']}")
            if value["isni"].startswith("http"):
                raise ValueError(
                    "ISNI identifiers must not start with 'http'. "
                    "Use only the identifier itself, e.g. '0000 0001 6785 333X'."
                )

    affiliations: list[Organization] = attrs.field(factory=list)
    """The affiliation of the creator, e.g. a university or research institute."""

    def to_zenodo(self) -> dict[str, Any]:
        """Convert the creator to a dictionary suitable for Zenodo."""
        result = {
            "person_or_org": {
                "type": "personal",
                "family_name": self.family_name,
                "given_name": self.given_name,
            }
        }
        if len(self.identifiers) > 0:
            result["person_or_org"]["identifiers"] = [
                {"scheme": key, "identifier": value} for key, value in self.identifiers.items()
            ]
        if len(self.affiliations) > 0:
            result["affiliations"] = [aff.to_zenodo() for aff in self.affiliations]
        return result


@attrs.define
class Access:
    """A subset of Invenio RDM Access.

    https://inveniordm.docs.cern.ch/reference/metadata/#access
    """

    record: str = attrs.field(
        default="public",
        validator=attrs.validators.in_(["public", "restricted"]),
    )
    """The access level of the record. Public means that the record is visible to everyone."""

    files: str = attrs.field(
        default="public",
        validator=attrs.validators.in_(["public", "restricted"]),
    )
    """The access level of the files. Public means that the files are visible to everyone."""

    def to_zenodo(self) -> dict[str, str]:
        """Convert the access configuration to a dictionary suitable for Zenodo."""
        return {
            "record": self.record,
            "files": self.files,
        }


@attrs.define
class Related:
    scheme: str = attrs.field(validator=attrs.validators.in_(IDENTIFIER_SCHEMES))
    """The scheme of the identifier, e.g. 'doi', 'arxiv'"""

    identifier: str = attrs.field()
    """The identifier itself, e.g. '10.5281/zenodo.1234567'"""

    @identifier.validator
    def _validate_identifier(self, attribute, value):
        """Validate the identifier."""
        if not IDENTIFIER_SCHEMES[self.scheme](value):
            raise ValueError(
                f"Invalid identifier for scheme {self.scheme}: {value}. "
                "Please check the identifier format."
            )

    relation_type: str = attrs.field(validator=attrs.validators.in_(RELATION_TYPES))

    resource_type: str | None = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.in_(RESOURCE_TYPES)),
    )

    def to_zenodo(self) -> dict[str, Any]:
        """Convert the related identifier to a dictionary suitable for Zenodo."""
        result = {
            "scheme": self.scheme,
            "identifier": self.identifier,
            "relation_type": {"id": self.relation_type},
        }
        if self.resource_type is not None:
            result["resource_type"] = {"id": self.resource_type}
        return result


@attrs.define
class Award:
    """A subset of Invenio RDM award information."""

    id: str | None = attrs.field(default=None)
    """The award ID, e.g. '123456'."""

    title: str | None = attrs.field(default=None)
    """The title of the award, e.g. 'Research Excellence Award'."""

    number: str | None = attrs.field(default=None)
    """The award number, e.g. 'RE-123456'."""

    identifiers: list[dict[str, str]] = attrs.field(
        factory=list,
    )
    """Identifiers of the award, e.g. a DOI or a funder ID. Keys must be lower case strings."""

    @identifiers.validator
    def _validate_identifiers(self, attribute, value):
        """Validate the identifiers."""
        if not isinstance(value, list):
            raise TypeError("Identifiers must be a list of dictionaries.")
        for item in value:
            if not isinstance(item, dict):
                raise TypeError("Each identifier must be a dictionary.")
            if len(item) != 1:
                raise ValueError("Each identifier must have exactly one key.")
            scheme, identifier = next(iter(item.items()))
            if scheme not in IDENTIFIER_SCHEMES:
                raise ValueError(f"Unknown identifier scheme: {scheme}")
            if not IDENTIFIER_SCHEMES[scheme](identifier):
                raise ValueError(f"Invalid identifier for scheme {scheme}: {identifier}")

    def __attrs_post_init__(self):
        if not ((self.id is None) ^ (self.title is None and self.number is None)):
            raise ValueError("Exactly one of 'id' or ('title', 'number') must be set for an award.")

    def to_zenodo(self) -> dict[str, Any]:
        """Convert the award information to a dictionary suitable for Zenodo."""
        result = {}
        if self.id is not None:
            result["id"] = self.id
        if self.title is not None:
            result["title"] = {"en": self.title}
        if self.number is not None:
            result["number"] = self.number
        if len(self.identifiers) > 0:
            result["identifiers"] = []
            for item in self.identifiers:
                if not isinstance(item, dict) or len(item) != 1:
                    raise ValueError("Each identifier must be a dictionary with exactly one key.")
                scheme, identifier = next(iter(item.items()))
                result["identifiers"].append({"scheme": scheme, "identifier": identifier})
        return result


@attrs.define
class Funding:
    """A subset of Invenio RDM funding information."""

    funder: Organization = attrs.field()
    """The funder of the project leading to the record, e.g. a research council or a foundation."""

    award: Award = attrs.field()
    """The award associated with the funding, if any."""

    def to_zenodo(self) -> dict[str, Any]:
        """Convert the funding information to a dictionary suitable for Zenodo."""
        return {
            "funder": self.funder.to_zenodo(),
            "award": self.award.to_zenodo(),
        }


def _convert_license(arg):
    if isinstance(arg, str):
        arg = [arg]
    return [lic.lower() for lic in arg]


@attrs.define
class Metadata:
    """A subset of Invenio RDM metadata.

    https://inveniordm.docs.cern.ch/reference/metadata/#metadata
    """

    title: str = attrs.field(validator=attrs.validators.min_len(1))
    version: str = attrs.field()
    license: str | list[str] = attrs.field(converter=_convert_license)
    resource_type: str = attrs.field(validator=attrs.validators.in_(RESOURCE_TYPES))
    copyright: str | None = attrs.field(default=None)
    publisher: str | None = attrs.field(default=None)
    keywords: list[str] = attrs.field(factory=list)
    creators: list[Creator] = attrs.field(factory=list)
    description: str | None = attrs.field(default=None)
    related: list[Related] = attrs.field(factory=list)
    funding: list[Funding] = attrs.field(factory=list)

    def to_zenodo(self) -> dict[str, Any]:
        """Convert the metadata to a dictionary suitable for Zenodo."""
        data = {
            "title": self.title,
            "version": self.version,
            "rights": [{"id": lic} for lic in self.license],
            "resource_type": {"id": self.resource_type},
            "creators": [creator.to_zenodo() for creator in self.creators],
            "description": self.description,
            "publication_date": datetime.date.today().isoformat(),
            "related_identifiers": [rel.to_zenodo() for rel in self.related],
            "funding": [fund.to_zenodo() for fund in self.funding],
        }
        if self.copyright is not None:
            data["copyright"] = self.copyright
        if self.publisher is not None:
            data["publisher"] = self.publisher
        if len(self.keywords) > 0:
            data["subjects"] = [{"subject": keyword} for keyword in self.keywords]
        return data


@attrs.define
class Config:
    """Configuration of the sync-zenodo script.

    An object of this class is created from data loaded from a local YAML config file.
    """

    path_record_id: Path = attrs.field(converter=Path)
    endpoint: str = attrs.field()
    path_token: str = attrs.field()
    metadata: Metadata = attrs.field()
    access: Access = attrs.field(default=Access())
    path_readme: str | None = attrs.field(default=None)
    paths: list[Path] = attrs.field(factory=list, converter=lambda paths: [Path(p) for p in paths])
    code_repository: str | None = attrs.field(default=None)

    def to_zenodo(self) -> dict[str, Any]:
        """Convert the configuration to a dictionary suitable for Zenodo."""
        data = {
            "access": self.access.to_zenodo(),
            "metadata": self.metadata.to_zenodo(),
            "files": {
                "enabled": len(self.paths) > 0,
                "order": [path.name for path in self.paths],
            },
        }
        if len(self.paths) > 0:
            data["files"]["default_preview"] = self.paths[0].name
        if self.code_repository is not None:
            data["custom_fields"] = {"code:codeRepository": self.code_repository}
        return data


@attrs.define
class ZenodoWrapper:
    """Python interface to a subset of the Zenodo API."""

    token: str = attrs.field()
    endpoint: str = attrs.field(default="https://sandbox.zenodo.org/api")
    verbose: bool = attrs.field(default=False)
    rest: RESTWrapper = attrs.field(init=False)

    @rest.default
    def _default_rest(self):
        """Set the default RESP Wrapper."""
        params = {"access_token": self.token, "Accept": "application/vnd.inveniordm.v1+json"}
        return RESTWrapper(self.endpoint, params, verbose=self.verbose)

    # Main API methods

    def create_new_record(self, config: Config) -> dict[str]:
        """Create a new record on Zenodo, which remains in draft until it is published manually."""
        return self.rest.post("records", json=config.to_zenodo())

    def get_record(self, rid: int) -> dict[str]:
        """Get an (un)published record with given id."""
        try:
            # If the dataset is in draft
            res = self.rest.get(f"records/{rid}/draft")
        except RESTError:
            # If the dataset is published
            res = self.rest.get(f"records/{rid}")
        return res

    def update_metadata(self, rid: int, config: Config) -> dict[str]:
        """Update the metadata of a record.

        This is applicable draft records and published records in edit mode.
        """
        return self.rest.put(f"records/{rid}/draft", json=config.to_zenodo())

    def edit_record(self, rid: int):
        """Put a published record into edit mode."""
        self.rest.post(f"records/{rid}/draft")

    def publish_record(self, rid: int):
        """Publish are draft record or a record in edit mode."""
        self.rest.post(f"records/{rid}/draft/actions/publish")

    def start_uploads(self, rid: int, paths: list[Path]):
        self.rest.post(f"records/{rid}/draft/files", json=[{"key": path.name} for path in paths])

    def upload_file(self, rid: int, path: Path):
        """Upload a file. The bucket must belong to a record in draft mode."""
        # Upload the file
        with open(path, "rb") as fh:
            self.rest.put(f"records/{rid}/draft/files/{path.name}/content", data=fh)
        # Commit it
        res = self.rest.post(f"records/{rid}/draft/files/{path.name}/commit")

        if not res["checksum"].startswith("md5:"):
            raise ZenodoError(f"Zenodo returned an unexpected checksum format: {res['checksum']}")

        # Check the checksum and file size
        if not res["size"] == path.size:
            raise ZenodoError(
                f"File size mismatch for {path.name}: "
                f"expected {path.size}, got {res['size']} bytes."
            )
        local_md5 = _compute_md5(path)
        if local_md5 != res["checksum"][4:]:
            raise ZenodoError(
                f"MD5 checksum mismatch for {path.name}, "
                f"local {local_md5}, online {res['checksum'][4:]}."
            )

    def delete_file(self, rid: int, name: str):
        """Delete a file. The file must belong to a record in draft mode."""
        self.rest.delete(f"records/{rid}/draft/files/{name}")

    def get_latest_version(self, parent_rid: int) -> dict[str] | None:
        """Get the latest version of a record."""
        return self.rest.get(f"records/{parent_rid}/versions")

    def create_new_version(self, rid: int) -> dict[str]:
        """Create a new version of a published record. The result is a draft record."""
        return self.rest.post(f"records/{rid}/versions")


def main(argv: list[str] | None = None):
    """Main program."""
    parser = argparse.ArgumentParser(
        prog="sync-zenodo",
        description="Synchronize a draft dataset on Zenodo with your local files.",
    )
    add_parser_args(parser)
    args = parser.parse_args(argv)
    sync_zenodo_tool(args)


def sync_zenodo_tool(args: argparse.Namespace) -> int:
    """Main program."""
    with open(args.config) as fh:
        data = yaml.safe_load(fh)
    if not isinstance(data.get("metadata").get("version"), str):
        raise TypeError(
            "The version in the YAML config file is missing or is not a string. "
            "Enclose the version in quotes to make it a string."
        )
    try:
        config = cattrs.structure(data, Config)
    except cattrs.ClassValidationError as exc:
        for line in cattrs.transform_error(exc, repr(args.config)):
            print(line)
        raise
    # Override the token path from the config file if the environment variable is set.
    path_token = getenv("REPREP_PATH_ZENODO_TOKEN")
    if path_token is not None:
        config.path_token = path_token
    if args.clean:
        clean_online(config, args.verbose)
    update_online(config, args.verbose)
    return 0


def sync_zenodo_subcommand(subparser: argparse.ArgumentParser) -> callable:
    """Create parser for command-line options."""
    parser = subparser.add_parser(
        "sync-zenodo",
        help="Synchronize a draft dataset on Zenodo with your local files.",
    )
    add_parser_args(parser)
    return sync_zenodo_tool


def add_parser_args(parser: argparse.ArgumentParser):
    """Define command-line arguments."""
    parser.add_argument("config", help="Configuration YAML file.")
    parser.add_argument(
        "--verbose",
        "-v",
        default=False,
        action="store_true",
        help="Show details of communication with Zenodo endpoint.",
    )
    parser.add_argument(
        "--clean",
        default=False,
        action="store_true",
        help="Remove all draft data sets before proceeding. "
        "This also deletes the record ID json file.",
    )


def clean_online(config: Config, verbose: bool):
    """Remove all draft data sets on Zenodo."""
    # (Try to) get the token.
    path_token = Path(config.path_token).expand()
    if not path_token.is_file():
        return
    with open(path_token) as fh:
        zenodo = ZenodoWrapper(fh.read().strip(), config.endpoint, verbose=verbose)

    # Get all records.
    response = zenodo.rest.get("user/records")
    for record in response["hits"]["hits"]:
        rid = record["id"]
        if record.get("submitted", False):
            print(f"Record {rid} is already published, skipping.")
            continue
        print(f"Deleting draft record {rid}.")
        try:
            zenodo.rest.delete(f"records/{rid}/draft")
        except RESTError as exc:
            print(f"Failed to delete record {rid}: {exc}")
    if config.path_record_id is not None:
        config.path_record_id.remove_p()


def update_online(config: Config, verbose: bool):
    """Make the online data set up to date with the local information."""
    # Amend inputs when this script is called in a StepUp workflow.
    paths_inp = list(config.paths)
    if config.path_readme is not None:
        paths_inp.append(config.path_readme)
    amend(inp=paths_inp)

    # If present, convert README Markdown file to HTML
    if config.path_readme is not None:
        with open(config.path_readme) as fh:
            description = fh.read()
        if config.path_readme.endswith(".md"):
            md = MarkdownIt()
            description = md.render(description)
        elif not config.path_readme.endswith(".html"):
            raise ZenodoError("The README file must be a Markdown (.md) or HTML (.html) file.")
        config.metadata.description = description

    # Run sanity check on paths: duplicate filenames not allowed.
    paths = {}
    for path in config.paths:
        path = Path(path)
        paths[path.name] = path
    if len(paths) != len(config.paths):
        raise ZenodoError(
            "Zenodo does not support directory layouts. Files must have different names."
        )

    # (Try to) get the token.
    path_token = Path(config.path_token).expand()
    if not path_token.is_file():
        return
    with open(path_token) as fh:
        zenodo = ZenodoWrapper(fh.read().strip(), config.endpoint, verbose=verbose)

    # Check if the parent record ID file exists, and load if it does.
    rid = None
    if config.path_record_id.exists():
        with open(config.path_record_id) as fh:
            try:
                rid = int(fh.read().strip())
            except ValueError as exc:
                raise ZenodoError(f"Invalid record ID in {config.path_record_id}: {exc}") from exc

    # Interact with Zenodo.
    if rid is None:
        # New record, when getting started with a dataset.
        record = _create_new(zenodo, config)
        with open(config.path_record_id, "w") as fh:
            print(record["id"], file=fh)
    else:
        # When a dataset exists, the actions depend on the current status of the record.
        record = zenodo.get_record(rid)
        # At the time of coding, the Zenodo API does not return the full metadata of the record.
        rid = record["id"]
        zenodo_version = record["metadata"]["version"]
        if record.get("submitted", False):
            cmpver = semver.compare(config.metadata.version, zenodo_version)
            if cmpver == 0:
                _check_record_md5(record, paths, config.metadata.version)
                # Unconditionally republish the metadata, because we cannot compare the metadata.
                _republish_metadata(zenodo, rid, config)
            elif cmpver > 0:
                record = _create_new_version(zenodo, rid, config)
                with open(config.path_record_id, "w") as fh:
                    print(record["id"], file=fh)
                _refresh_files(zenodo, record, paths, config.metadata.version)
            else:
                raise ZenodoError(
                    f"The online version ({zenodo_version}) is newer "
                    f"than the local one ({config.metadata.version})."
                )
        else:
            _refresh_files(zenodo, record, paths, config.metadata.version)
            # Unconditionally update the metadata, because we cannot compare the metadata.
            zenodo.update_metadata(rid, config)


def _create_new(zenodo: ZenodoWrapper, config: Config) -> dict[str, Any]:
    """Create a new record on Zenodo."""
    record = zenodo.create_new_record(config)

    # Declare the files to be uploaded.
    zenodo.start_uploads(record["id"], config.paths)

    # Actual uploads, one by one.
    for path in config.paths:
        print(f"Uploading {path}")
        zenodo.upload_file(record["id"], path)
    return record


def _compute_md5(path: str) -> bool:
    """Compute the MD5 sum of a file and compare to the given checksum."""
    with open(path, "rb") as fh:
        return hashlib.file_digest(fh, hashlib.md5).hexdigest()


def _check_record_md5(record: dict[str], paths: dict[str, Path], version: str):
    """Sanity check of MD5 hashes received from Zenodo"""
    for file in record["files"]:
        if file["key"] not in paths:
            raise ZenodoError(
                f"File {file['key']} exists online but not locally ({version}, published)"
            )
        path = paths[file["key"]]
        local_md5 = _compute_md5(path)
        if local_md5 != file["checksum"][4:]:
            raise ValueError(
                f"MD5 Checksum mismatch for {path} ({version}, published), "
                f"local: {local_md5}, online: {file['checksum'][4:]}"
            )
    online_names = {file["key"] for file in record["files"]}
    for name in paths:
        if name not in online_names:
            raise OSError(f"File {name} exists locally but not online ({version}, published)")


def _republish_metadata(zenodo: ZenodoWrapper, rid: int, config: Config):
    """Put a record in edit mode, update the metadata and publish again."""
    print(f"Editing metadata and publishing same version ({config.metadata.version}) again.")
    zenodo.edit_record(rid)
    zenodo.update_metadata(rid, config)
    zenodo.publish_record(rid)


def _create_new_version(zenodo: ZenodoWrapper, rid: int, config: Config) -> dict[str]:
    """Create a new version of the dataset and refresh the metadata."""
    print(f"Creating a new version ({config.metadata.version})")
    record = zenodo.create_new_version(rid)
    return zenodo.update_metadata(record["id"], config)


def _refresh_files(
    zenodo: ZenodoWrapper, record: dict[str], paths: dict[str, Path], version: str
) -> bool:
    """Refresh the online files.

    This function only uploads files that do not exist online yet or have changed locally.
    Files removed from the local YAML config will also be removed online.
    """
    changed = False
    for file in record["files"]:
        if not file["checksum"].startswith("md5:"):
            raise ZenodoError(f"Zenodo returned an unexpected checksum format: {file['checksum']}")
        if file["key"] not in paths:
            print(f"Deleting {file['key']} ({version}, draft)")
            zenodo.delete_file(record["id"], file["key"])
            changed = True
        else:
            path = paths[file["key"]]
            local_md5 = _compute_md5(path)
            if local_md5 != file["checksum"][4:]:
                print(f"Replacing {path} ({version}, draft)")
                zenodo.delete_file(record["id"], file["key"])
                zenodo.start_uploads(record["id"], [path])
                zenodo.upload_file(record["id"], path)
                changed = True
    online_names = {file["key"] for file in record["files"]}
    for name, path in paths.items():
        if name not in online_names:
            print(f"Uploading {path} ({version}, draft)")
            zenodo.start_uploads(record["id"], [path])
            zenodo.upload_file(record["id"], path)
            changed = True
    return changed


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
r"""Identification of errors in LaTeX logs."""

import re
import sys

import attrs


@attrs.define
class ErrorInfo:
    program: str = attrs.field(validator=attrs.validators.instance_of(str))
    src: str | None = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )
    message: str | None = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
    )

    def print(self, log: str | None = None):
        print(f"\033[1;31;40m{self.program} ERROR\033[0;0m", file=sys.stderr)
        if log is not None:
            print(f"\033[1;35;40mLog file:\033[0;0m {log}", file=sys.stderr)
        if self.src is not None:
            print(f"\033[1;35;40mSource file:\033[0;0m {self.src}", file=sys.stderr)
        if self.message is not None:
            print(self.message, file=sys.stderr)


DEFAULT_MESSAGE = """\
> The error message could not be isolated from the file {path}.
> You can open the file {path} in a text editor and locate the error manually.
>
> Please open a new issue with the file {path} attached,
> which will help improve the script to detect the error message:
> https://github.com/reproducible-reporting/stepup-reprep/issues
>
> Thank you very much!
"""

MESSAGE_SUFFIX = """
> If the above extract from the log file can be improved,
> open a new issue with the file {path} attached:
> https://github.com/reproducible-reporting/stepup-reprep/issues
"""


@attrs.define
class LatexSourceStack:
    stack: list[str] = attrs.field(init=False, default=attrs.Factory(list))
    unfinished: str | None = attrs.field(init=False, default=None)
    unmatched: bool = attrs.field(init=False, default=False)

    @property
    def current(self) -> str:
        """The current file to which the error message belongs."""
        if len(self.stack) == 0:
            return "(could not detect source file)"
        return self.stack[-1]

    def feed(self, line: str):
        # Check if we need to anticipate line wrapping
        full = len(line) == 80
        if full:
            # Some exceptions: guess when 80-char lines end exactly with a filename.
            # This is fragile, but LaTeX log files are just a mess to parse.
            for end in ".tex\n", ".sty\n", ".cls\n", ".def\n", ".cfg\n", ".clo\n":
                if line.endswith(end):
                    full = False
                    break

        # Continue from previous line if needed
        if self.unfinished is not None:
            line = self.unfinished + line
            self.unfinished = None

        if full:
            self.unfinished = line[:-1]
            return

        # Update to stack
        brackets = re.findall(r"\((?:(?:\./|\.\./|/)[-_./a-zA-Z0-9]+)?|\)", line)
        for bracket in brackets:
            if bracket == ")":
                if len(self.stack) == 0:
                    self.unmatched = True
                else:
                    del self.stack[-1]
            else:
                if not bracket.startswith("("):
                    raise AssertionError("Inconsistent parenthesis logic in LatexSourceStack")
                self.stack.append(bracket[1:])


def parse_latex_log(path_log: str) -> ErrorInfo | None:
    """Parse a LaTeX log file.

    Parameters
    ----------
    path_log
        The log file

    Returns
    -------
    error_info
        Structured info for printing error, or None
    """
    lss = LatexSourceStack()
    src = "(could not detect source file)"
    record = False
    found_line = False
    recorded = []

    # LaTeX log files may have encoding errors, so such errors must be ignored.
    with open(path_log, errors="ignore") as fh:
        for line in fh.readlines():
            if record:
                recorded.append(line.rstrip())
                if recorded[-1].strip() == "":
                    record = False
                    if found_line:
                        break
            if line.startswith("!"):
                if not record:
                    recorded.append(line.rstrip())
                record = True
                src = lss.current
            elif line.startswith("l."):
                if not record:
                    recorded.append(line.rstrip())
                record = True
                found_line = True
            else:
                lss.feed(line)

    if len(recorded) > 0:
        message = "\n".join(recorded) + MESSAGE_SUFFIX.format(path=path_log)
    else:
        message = DEFAULT_MESSAGE.format(path=path_log)
    if lss.unmatched:
        message += "> [warning: unmatched closing parenthesis]\n"
    return ErrorInfo("LaTeX", src, message=message)


def update_last_src(line, last_src):
    if not ("(" in line or ")" in line or line.startswith("/")):
        last_src = line
    return last_src



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Remove trailer ID and flaky metadata to make PDFs reproducible."""

import argparse
import shutil
import sys
import tempfile

import fitz
from path import Path

__all__ = ("pdf_normalize",)


def main(argv: list[str] | None = None):
    """Main program."""
    pdf_normalize(parse_args(argv).path_pdf)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(prog="rr-normalize-pdf", description="Normalize a PDF file.")
    parser.add_argument("path_pdf", help="The pdf to be normalized (in place).")
    return parser.parse_args(argv)


def pdf_normalize(path_pdf: str):
    """Replace a PDF file by its normalized equivalent. This helps making PDFs reproducible."""
    if not path_pdf.endswith(".pdf"):
        raise ValueError(f"The input must have a `.pdf` extension, got: {path_pdf}")
    pdf = fitz.open(path_pdf)
    pdf.set_metadata({})
    pdf.del_xml_metadata()
    pdf.xref_set_key(-1, "ID", "null")
    pdf.scrub()
    with tempfile.TemporaryDirectory(suffix="rr-normalize-pdf", prefix="rr") as dn:
        path_out = Path(dn) / "out.pdf"
        pdf.save(path_out, garbage=4, deflate=True, no_new_id=True)
        pdf.close()
        shutil.copy(path_out, path_pdf)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""RepRep Wrapper for git commands that take the current commit id as a dependency."""

import argparse
import shlex
import sys

from path import Path

from stepup.core.api import amend
from stepup.core.utils import myrelpath
from stepup.core.worker import WorkThread


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        prog="wrap_git", description="Run git commands that depend on the current commit id."
    )
    parser.add_argument(
        "--out",
        type=Path,
        help="Output file to write the command to. "
        "If not specified, the command is printed to stdout.",
    )
    parser.add_argument(
        "git_args",
        nargs="*",
        type=str,
        help="The git command to run.",
    )
    return parser.parse_args(argv)


def main(argv: list[str] | None = None, work_thread: WorkThread | None = None) -> None:
    """Main program."""
    args = parse_args(argv)
    if work_thread is None:
        work_thread = WorkThread("stub")

    # Determine the git root directory.
    return_code, git_root, _ = work_thread.runsh("git rev-parse --show-toplevel")
    if return_code != 0:
        raise RuntimeError("Failed to determine git root directory.")
    git_root = myrelpath(git_root.strip())

    # Mark the HEAD file as an input dependency.
    head_path = git_root / ".git" / "HEAD"
    amend(inp=head_path)
    with open(head_path) as fh:
        head_content = fh.read().strip()
        if not head_content.startswith("ref: "):
            raise ValueError("HEAD is not a symbolic reference, cannot determine commit id.")
    ref_path = head_content[5:].strip()

    # Mark the ref path as an input dependency.
    amend(inp=git_root / ".git" / ref_path)

    # Run the git command.
    command = "GIT_PAGER=cat " + shlex.join(args.git_args)
    if args.out is not None:
        command = f"{command} > {shlex.quote(args.out)}"
    work_thread.runsh_verbose(command)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""StepUp reprep package."""



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Create inventory files.

There are two kinds of inventory files:

- ``inventory.def`` defines rules to include and exclude files to be archived.
- ``inventory.txt`` is the result of a processing a ``inventory.def`` file with this script.
  It contains every individual file, together with its size in bytes and its blake2b hexdigest.
"""

import argparse
import contextlib
import shlex
import sqlite3
import subprocess
import sys
from collections.abc import Collection

from path import Path

from stepup.core.api import amend
from stepup.core.file import FileState
from stepup.core.nglob import NGlobMulti

from .inventory import format_summary, get_summary

__all__ = ("main", "write_inventory")


def main(argv: list[str] | None = None):
    """Main program."""
    parser = argparse.ArgumentParser(
        prog="rr-make-inventory", description="Make an inventory.txt file."
    )
    add_parser_args(parser)
    args = parser.parse_args(argv)
    make_inventory(args)


def add_parser_args(parser: argparse.ArgumentParser):
    """Define command-line arguments."""
    parser.add_argument(
        "paths",
        nargs="*",
        help="File to include in the inventory. "
        "(Added after porcessing the inventory definition if any.)",
    )
    parser.add_argument("-i", "--inventory-def", help="An inventory definition file.", default=None)
    parser.add_argument("-o", "--inventory-txt", help="An inventory output file.", default=None)


def make_subcommand(subparser: argparse.ArgumentParser) -> callable:
    """Create a subcommand for the command line interface."""
    parser = subparser.add_parser(
        "make-inventory",
        help="Create an inventory.txt file.",
    )
    add_parser_args(parser)
    return make_tool


def make_tool(args: argparse.Namespace) -> int:
    """Create an inventory.txt file."""
    make_inventory(args)
    return 0


def make_inventory(args: argparse.Namespace):
    # Check arguments
    if args.inventory_def is None:
        if len(args.paths) == 0:
            raise ValueError("At least inventory.def or a list of files are needed as input.")
        if args.inventory_txt is None:
            raise ValueError("Whithout -i inventory.def, the -o option is no longer optional.")
    else:
        if not args.inventory_def.endswith(".def"):
            raise ValueError("The inventory defintion file must end with .def")
        if args.inventory_txt is None:
            args.inventory_txt = args.inventory_def[:-4] + ".txt"
        elif Path(args.inventory_txt).parent != Path(args.inventory_def).parent:
            raise ValueError(
                "The inventory_def and inventory_txt files must have the same parent directory."
            )
    if not args.inventory_txt.endswith(".txt"):
        raise ValueError("The inventory output file must end with .txt")

    # Collect the complete list of files.
    path_inventory_txt = Path(args.inventory_txt)
    root = path_inventory_txt.parent.normpath()
    if args.inventory_def is None:
        paths = set()
    else:
        with open(args.inventory_def) as fh:
            lines = fh.readlines()
        with contextlib.chdir(root):
            paths = {root / path for path in parse_inventory_def(lines)}
    paths.update(args.paths)
    write_inventory(path_inventory_txt, sorted(paths))


def get_file_list_nglob(i: int, args: list[str]) -> Collection[Path]:
    if len(args) == 0:
        raise ValueError(
            f"Error on line {i} of the inventory definition: include or exclude has no arguments."
        )
    ngm = NGlobMulti.from_patterns(args)
    ngm.glob()
    return ngm.files()


def get_file_list_git(i: int, args: list[str]) -> Collection[Path]:
    cp = subprocess.run(
        ["git", "ls-files", *args],
        stdin=subprocess.DEVNULL,
        capture_output=True,
        check=True,
        encoding="utf-8",
    )
    return [Path(line.strip()) for line in cp.stdout.splitlines()]


def get_file_list_workflow(i: int, args: list[str]) -> Collection[Path]:
    """Get a list of files from a StepUp graph.db workflow file.

    There must be at least two arguments: the state and one or more patterns of graph.db files.
    """
    if len(args) < 2:
        raise ValueError(
            f"Error on line {i} of the inventory definition: Expecting at least two arguments."
        )
    state = FileState[args[0]]
    ngm = NGlobMulti.from_patterns(args[1:])
    ngm.glob()
    paths_graph_db = ngm.files()
    if len(paths_graph_db) == 0:
        raise ValueError(
            f"Error on line {i} of the inventory definition: no matching graph.db workflow files."
        )
    paths = set()
    for path_graph_db in paths_graph_db:
        if path_graph_db.parent.name != ".stepup":
            raise ValueError("A graph.db file must be in a .stepup directory.")
        root = path_graph_db.parent.parent
        con = sqlite3.connect(f"file:{path_graph_db}?mode=ro", uri=True)
        sql = "SELECT label FROM node JOIN file ON node.i = file.node WHERE state = ?"
        for (path,) in con.execute(sql, (state.value,)):
            if not path.endswith("/"):
                paths.add(root / path)
    return paths


FILE_LIST_SOURCES = {
    "": get_file_list_nglob,
    "-git": get_file_list_git,
    "-workflow": get_file_list_workflow,
}


def parse_inventory_def(lines: list[str], paths: list[str] | None = None) -> set[str]:
    """Process lines from an inventory definition file.

    Parameters
    ----------
    lines
        A list of single lines from an inventory definition file.
    paths
        A list of paths to use as a starting point, if any.
        When not given, this function starts from an empty list.

    Returns
    -------
    paths
        A set of paths obtained by processing the include and exclude commands.
        When an initial paths list is given, it is not altered.
    """
    paths = set() if paths is None else set(paths)
    for i, line in enumerate(lines):
        words = shlex.split(line, comments=True)
        if len(words) == 0:
            continue
        command = words[0]
        args = words[1:]
        action = command[:7].lower()
        if action not in ["include", "exclude"]:
            raise ValueError(f"Line {i} does not start with include or exclude: {line}")
        source = command[7:].lower()
        file_list_function = FILE_LIST_SOURCES.get(source)
        if file_list_function is None:
            raise ValueError(f"Unsupported command on line {i}: {command}")
        new_paths = file_list_function(i, args)
        new_paths = [path for path in new_paths if not path.is_dir()]
        if len(new_paths) == 0:
            raise ValueError(f"Line {i} matches no paths: {line}")
        if action == "include":
            paths.update(new_paths)
        else:
            paths.difference_update(new_paths)
    return paths


def write_inventory(path_txt: str, paths: Collection[str], do_amend: bool = True):
    """Write an inventory file.

    Parameters
    ----------
    path_txt
        The location of the inventory file to be written.
    paths
        A collection of paths to be included in the inventory.
        These must be paths relative to the current working directory.
        They will be written to the inventory file
        as paths relative to the parent of the inventory file.
    """
    # Amend all paths included in the inventory file as inputs.
    # This is needed to ensure that the inventory is rebuilt when the paths change.
    # Other actions calling this function may not want this,
    # because they already take care of file dependencies and amendments
    # may then cause cyclic dependencies.
    if do_amend:
        inp_paths = []
        for path in paths:
            path = Path(path)
            if path.is_dir() and not path.endswith("/"):
                path = path / ""
            inp_paths.append(path)
        amend(inp=inp_paths)

    # Write the inventory file.
    path_txt = Path(path_txt)
    root = path_txt.parent.normpath()
    with open(path_txt, "w") as fh:
        for path in paths:
            print(format_summary(get_summary(path, root)), file=fh)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
r"""Identification of errors in BibTeX logs."""

from .latex_log import DEFAULT_MESSAGE, ErrorInfo


def parse_bibtex_log(path_blg: str) -> ErrorInfo | None:
    """Parse a BibTeX log file.

    Parameters
    ----------
    path_blg
        The blg file.

    Returns
    -------
    error_info
        Structured info for printing error, or None
    """
    last_src = "(could not detect source file)"
    error = False
    recorded = []
    with open(path_blg, errors="ignore") as fh:
        for line in fh.readlines():
            if "---" in line and "file " in line:
                last_src = line.rsplit(maxsplit=1)[-1]
                recorded = []
            recorded.append(line[:-1])
            if line.startswith("I'm skipping whatever remains"):
                error = True
                break
            if line.startswith(r"I found no \bibstyle command"):
                last_src = line.split()[-1]
                recorded = [line[:-1]]
                error = True
                break

    message = "\n".join(recorded) if error else DEFAULT_MESSAGE.format(path=path_blg)
    return ErrorInfo("BibTeX", last_src, message=message)



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Script to insert notes pages at every even page in a PDF."""

import argparse
import sys

import fitz

__all__ = ("add_notes_pdf",)


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    add_notes_pdf(args.path_src, args.path_notes, args.path_dst)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-add-notes-pdf",
        description="Add a notes page at every even page of a PDF file.",
    )
    parser.add_argument("path_src", help="The source pdf to which notes should be added.")
    parser.add_argument("path_notes", help="The pdf with the notes page(s).")
    parser.add_argument("path_dst", help="The output pdf.")
    return parser.parse_args(argv)


def add_notes_pdf(path_src: str, path_notes: str, path_dst: str):
    """Insert notes pages at every even page."""
    for path_pdf in path_src, path_notes, path_dst:
        if not path_pdf.endswith(".pdf"):
            raise ValueError(f"All arguments must have a `.pdf` extension, got: {path_pdf}")
    src = fitz.open(path_src)
    # See https://github.com/pymupdf/PyMuPDF/issues/3635
    src.scrub()
    notes = fitz.open(path_notes)

    # Create the combined PDF
    dst = fitz.open()
    for isrc in range(len(src)):
        final = isrc == len(src) - 1
        dst.insert_pdf(src, from_page=isrc, to_page=isrc, final=final)
        inotes = isrc % len(notes)
        dst.insert_pdf(notes, from_page=inotes, to_page=inotes, final=final)

    # Strip metadata for reproducibility and save
    dst.set_metadata({})
    dst.del_xml_metadata()
    dst.xref_set_key(-1, "ID", "null")
    dst.scrub()
    dst.save(path_dst, garbage=4, deflate=True, no_new_id=True)

    dst.close()
    src.close()
    notes.close()


if __name__ == "__main__":
    sys.exit(main())



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
"""Sanitize BibTeX files."""

import argparse
import enum
import os
import re
import shutil
import sys
import tempfile
from collections.abc import Collection

import attrs
import cattrs
import pybtex.database
import yaml
from path import Path
from pyiso4.ltwa import Abbreviate
from pylatexenc.latex2text import LatexNodes2Text
from pylatexenc.latexencode import RULE_REGEX, UnicodeToLatexConversionRule, UnicodeToLatexEncoder

from stepup.core.hash import compute_file_digest

__all__ = ("BibsaneConfig", "main")


L2U = LatexNodes2Text().latex_to_text
CONVERSION_RULES = [
    # Custom LaTeX representation rules
    UnicodeToLatexConversionRule(
        RULE_REGEX,
        [
            # Use -- and --- for dashes when it is reasonable.
            (re.compile("—(?![-–])"), "---"),  # noqa: RUF001
            (re.compile("–(?![-—])"), "--"),  # noqa: RUF001
        ],
    ),
    # Add the default rules
    "defaults",
]
U2L = UnicodeToLatexEncoder(conversion_rules=CONVERSION_RULES).unicode_to_latex


@enum.unique
class DuplicatePolicy(enum.Enum):
    """The three policies for duplicate entries."""

    FAIL = "fail"
    MERGE = "merge"
    IGNORE = "ignore"


@enum.unique
class FieldPolicy(enum.Enum):
    """The two policies for bibliography fields."""

    MUST = "must"
    MAY = "may"


@attrs.define
class BibsaneConfig:
    """The configuration object controlling BibSane behavior.

    Note that the settings default to the most permissive and least invasive ones.
    We recommend the opposite settings, but you have to switch knowingly in the config file.
    """

    root: Path = attrs.field()
    """The parent directory of the configuration file."""

    duplicate_key: DuplicatePolicy = attrs.field(default=DuplicatePolicy.IGNORE)
    """The policy for duplicate BibTeX keys: fail, merge or ignore."""

    duplicate_doi: DuplicatePolicy = attrs.field(default=DuplicatePolicy.IGNORE)
    """The policy for duplicate DOIs: fail, merge or ignore."""

    drop_entry_types: list[str] = attrs.field(default=attrs.Factory(list))
    """The entry types to drop from the BibTeX database."""

    normalize_doi: bool = attrs.field(default=False)
    """Set to `True` to normalize the DOIs in the entries."""

    normalize_whitespace: bool = attrs.field(default=False)
    """Set to `True` to normalize the whitespace in the field values."""

    fix_page_double_hyphen: bool = attrs.field(default=False)
    """Set to `True` to fix the page ranges for which no double hyphen is used."""

    abbreviate_journals: bool = attrs.field(default=False)
    """Set to `True` to abbreviate journal names using ISO4."""

    custom_abbreviations: dict[str, str] = attrs.field(factory=dict)
    """Custom journal abbreviations.

    By default, pyiso4 is used to abbreviate journal names.
    The custom abbreviations can override those provided by pyiso4.
    """

    brace_title_words: bool = attrs.field(default=False)
    """Recognize words in titles that should be wrapped in braces.

    This is performed late, after encoding to LaTeX, to avoid that the braces
    are represented as `\\{` and `\\}`.
    """

    sort: bool = attrs.field(default=False)
    """Set to `True` to sort the entries by year and first author.

    The sort key is `{year}{first author lowercase last name}`.
    """

    citation_policies: dict[str, dict[str, FieldPolicy]] = attrs.field(default=attrs.Factory(dict))
    """The field policies (must or may) for each entry type."""

    @classmethod
    def from_file(cls, fn_yaml: str):
        """Instantiate a configuration from a YAML config file."""
        if fn_yaml is None:
            config = cls(os.getcwd())
        else:
            with open(fn_yaml) as f:
                data = yaml.safe_load(f)
                if data is None:
                    data = {}
                elif not isinstance(data, dict):
                    raise ValueError(
                        f"Invalid BibSane config file: expected a mapping at top level.({fn_yaml})"
                    )
                data.setdefault("root", os.path.dirname(fn_yaml))
                config = cattrs.structure(data, cls)
        return config


RETURN_CODE_SUCCESS = 0
RETURN_CODE_CHANGED = 1
RETURN_CODE_BROKEN = 2


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)

    # Load the bib file.
    print("📂 Load", args.bib)
    entries = collect_entries(args.bib)
    print(f"    Found {len(entries)} BibTeX entries")

    # Check for duplicate keys
    # TODO: This is currently pointless because pybtex already fails early on duplicate keys.
    if args.config.duplicate_key == DuplicatePolicy.FAIL:
        print("🔨 Checking for duplicate BibTeX entry keys")
        if not check_duplicate_keys(entries):
            print("    ❌ Stop early due to duplicate BibTeX entry keys")
            return RETURN_CODE_BROKEN

    # Check for duplicate DOIs
    if args.config.duplicate_doi == DuplicatePolicy.FAIL:
        print("🔨 Checking for duplicate DOIs")
        if not check_duplicate_dois(entries):
            print("    ❌ Stop early due to duplicate DOIs")
            return RETURN_CODE_BROKEN

    # Load the aux file.
    if args.aux is not None:
        if not args.aux.endswith(".aux"):
            print("    ❌ Aux file has no extension .aux:", args.aux)
            return RETURN_CODE_BROKEN
        print("📂 Load", args.aux)
        citations = parse_aux(args.aux)
        print(f"    Found {len(citations)} citations")
        citations = set(citations)
        print(f"    Found {len(citations)} unique citations")
        if len(citations) == 0:
            print("    ❓ Ignored aux file because it lacks citations.")
        else:
            # Drop unused and check for missing
            print("🔨 Checking unused and missing citations")
            bibdata_complete = check_citations(entries, citations)
            print(f"    Found {len(entries)} used BibTeX entries")
            if not bibdata_complete:
                print("    ❌ Stop early due to missing citations")
                return RETURN_CODE_BROKEN

    # Drop irrelevant entry types
    if len(args.config.drop_entry_types) > 0:
        print("🔨 Drop irrelevant entry types")
        drop_entry_types(entries, args.config.drop_entry_types)
        print(f"    {len(entries)} BibTeX entries left")

    # The default return code, assuming some changes are made but no errors found.
    retcode = RETURN_CODE_CHANGED

    # Clean entries
    if len(args.config.citation_policies) > 0:
        print("🔨 Apply and check citation policies")
        if not clean_entries(entries, args.config.citation_policies):
            retcode = RETURN_CODE_BROKEN

    # Check for potential problems that cannot be fixed automatically, not optional.
    # TODO: This is currently pointless because pybtex already fails early on duplicate keys.
    print("🔨 Check for potential mistakes in BibTeX keys")
    if not case_consistent_keys(entries):
        retcode = RETURN_CODE_BROKEN

    # Normalize the DOIs (lowercase and remove prefix)
    if args.config.normalize_doi:
        print("🔨 Normalize DOIs")
        valid_dois = normalize_doi(entries)
        if not valid_dois:
            retcode = RETURN_CODE_BROKEN

    # Remove redundant whitespace
    # TODO: This is mostly redundant because pybtex already normalizes whitespace.
    if args.config.normalize_whitespace:
        print("🔨 Normalize whitespace")
        normalize_whitespace(entries)

    # Fix page double hyphen
    if args.config.fix_page_double_hyphen:
        print("🔨 Fix double hyphen in page ranges")
        if not fix_page_double_hyphen(entries):
            retcode = RETURN_CODE_BROKEN

    # Abbreviate journal names
    if args.config.abbreviate_journals:
        print("🔨 Abbreviate journal names")
        abbreviate_journal_iso(entries, args.config.custom_abbreviations)

    # Merge entries
    if args.config.duplicate_key == DuplicatePolicy.MERGE:
        # TODO: This is currently pointless because pybtex already fails early on duplicate keys.
        print("🔨 Merge references by BibTeX key")
        if merge_entries(entries, KEY):
            retcode = RETURN_CODE_BROKEN
        print(f"    {len(entries)} entries left")
    if args.config.duplicate_doi == DuplicatePolicy.MERGE:
        print("🔨 Merge references by DOI")
        if merge_entries(entries, "doi"):
            retcode = RETURN_CODE_BROKEN
        print(f"    {len(entries)} entries left")

    # Sort entries
    if args.config.sort:
        print("🔨 Sort by Year + Author + Title")
        sort_entries(entries)

    # Overwrite if needed.
    fn_out = args.bib if args.out is None else args.out
    retcode = write_output(entries, fn_out, args.config.brace_title_words, retcode)
    if args.out is not None and retcode == RETURN_CODE_CHANGED:
        retcode = RETURN_CODE_SUCCESS
    return retcode


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-bibsane",
        description="Sanitize and clean up a BibTeX file.",
    )
    parser.add_argument("bib", help="The BibTeX file to check and clean up.")
    parser.add_argument(
        "-a",
        "--aux",
        help="The LaTeX aux file of your document. If given, unused entries will be dropped.",
    )
    parser.add_argument("-c", "--config", help="An optional configuration file")
    parser.add_argument(
        "--out",
        help="Output path for the cleaned up BibTeX file. "
        "If not given, rr-bibsane will be overwrite the given file with the cleaned up version "
        "(if any changes were needed). "
        "This will cause StepUp to drain the scheduler so you can inspect the changes and rebuild.",
    )
    args = parser.parse_args(argv)
    args.config = BibsaneConfig.from_file(args.config)
    return args


class BibsaneMagicConstants(enum.Enum):
    """Magic constants for BibSane."""

    BIB_ENTRY_TYPE = enum.auto()
    BIB_KEY = enum.auto()


# Short and convenient dictionary keys for special BibTeX entry fields.
ETYPE = BibsaneMagicConstants.BIB_ENTRY_TYPE
KEY = BibsaneMagicConstants.BIB_KEY


def collect_entries(fn_bib: str) -> list[dict]:
    """Collect entries from the BibTeX files into standard Python data structures."""
    entries = []
    lib = pybtex.database.parse_file(fn_bib)
    for pyb_entry in lib.entries.values():
        # Convert to a simple dictionary, to enforce modularity between bibsane and pybtex.
        # At this stage, all keys are lowercased and LaTeX is decoded with pylatexenc
        entry = {
            ETYPE: pyb_entry.type.lower(),
            KEY: pyb_entry.key,
        }
        for key, value in pyb_entry.fields.items():
            entry[key.lower()] = L2U(value)
        for role, names in pyb_entry.persons.items():
            entry[role.lower()] = " and ".join(L2U(str(name)) for name in names)
        entries.append(entry)
    return entries


def check_duplicate_keys(entries: list[dict]) -> bool:
    """Check for duplicate BibTeX entry keys."""
    seen_ids = set()
    valid = True
    for entry in entries:
        if entry[KEY] in seen_ids:
            print(f"    ‼️ Duplicate BibTeX entry: {entry[KEY]}")
            valid = False
        seen_ids.add(entry[KEY])
    return valid


def check_duplicate_dois(entries: list[dict]) -> bool:
    """Check for duplicate DOIs."""
    seen_dois = set()
    valid = True
    for entry in entries:
        doi = entry.get("doi")
        if doi is not None:
            if doi in seen_dois:
                print(f"    ‼️ Duplicate DOI: {doi}")
                valid = False
            seen_dois.add(doi)
    return valid


# List of citations to ignore, which are added by some LaTeX templates,
# but which are not correctly parsed by python-bibtexparser.
# Related issue: https://github.com/sciunto-org/python-bibtexparser/issues/384
IGNORED_CITATIONS = {"REVTEX41Control", "achemso-control"}


def parse_aux(fn_aux: str) -> list[str]:
    """Parse the relevant parts of a LaTeX aux file."""
    citations = []
    with open(fn_aux) as f:
        for line in f:
            parse_aux_line("citation", line, citations)
    # Filter out bogus citations
    return [citation for citation in citations if citation not in IGNORED_CITATIONS]


def parse_aux_line(prefix: str, line: str, words: list[str]):
    """Parse a (simple) line from a LaTeX aux file."""
    if line.startswith(rf"\{prefix}{{"):
        if not (line.endswith("}\n") and line.count("{") == 1 and line.count("}") == 1):
            print("    🤕 Cannot parse aux line:", line.strip())
            return
        words.extend(line[line.find("{") + 1 : -2].split(","))


def check_citations(entries: list[dict], citations: Collection[str]) -> bool:
    """Drop unused citations and complain about missing ones."""
    # Check for undefined references
    defined = {entry[KEY] for entry in entries}
    valid = True
    for citation in citations:
        if citation not in defined:
            print("    💀 Missing reference:", citation)
            valid = False

    # Drop unused entries
    result = []
    for entry in entries:
        if entry[KEY] not in citations:
            print("    🧹 Dropping unused key:", entry[KEY])
            continue
        result.append(entry)
    entries[:] = result
    return valid


def drop_entry_types(entries: list[dict], drop: Collection[str]):
    """Drop entries of the given types."""
    result = []
    for entry in entries:
        if entry[ETYPE] in drop:
            print("    🧹 Dropping irrelevant entry type:", entry[ETYPE])
            continue
        result.append(entry)
    entries[:] = result


def clean_entries(
    entries: list[dict], citation_policies: dict[str, dict[str, FieldPolicy]]
) -> bool:
    """Clean the irrelevant fields in each entry and complain about missing ones."""
    valid = True
    for old_entry in entries:
        etype = old_entry[ETYPE]
        key = old_entry[KEY]
        new_entry = {ETYPE: etype, KEY: key}
        if "bibsane" in old_entry:
            etype = old_entry.pop("bibsane")
            new_entry["bibsane"] = etype
        entry_policy = citation_policies.get(etype)
        if entry_policy is None:
            print(f"    🤔 {key}: @{etype} is not configured")
            valid = False
            continue
        for field, policy in entry_policy.items():
            if policy == FieldPolicy.MUST:
                if field not in old_entry:
                    print(f"    🫥 {key}: @{etype} missing field {field}")
                    valid = False
                else:
                    new_entry[field] = old_entry.pop(field)
            else:
                assert policy == FieldPolicy.MAY
                if field in old_entry:
                    new_entry[field] = old_entry.pop(field)
        if len(old_entry) > 0:
            for field in old_entry:
                if field not in (ETYPE, KEY):
                    print(f"    💨 {key}: @{etype} discarding field {field}")
        old_entry.clear()
        old_entry.update(new_entry)
    return valid


def case_consistent_keys(entries: list[dict]) -> bool:
    """Detect potential mistakes in the BibTeX entry keys."""
    id_case_map = {}
    for entry in entries:
        id_case_map.setdefault(entry[KEY].lower(), []).append(entry[KEY])

    valid = True
    for groups in id_case_map.values():
        if len(groups) > 1:
            print("    👻 BibTeX entry keys that only differ by case:", " ".join(groups))
            valid = False
    return valid


DOI_PROXIES = [
    "https://doi.org/",
    "http://doi.org/",
    "http://dx.doi.org/",
    "https://dx.doi.org/",
    "doi:",
]


def normalize_doi(entries: list[dict]) -> bool:
    """Normalize the DOIs in the entries."""
    valid = True
    for entry in entries:
        doi = entry.get("doi")
        if doi is not None:
            doi = doi.lower()
            for proxy in DOI_PROXIES:
                if doi.startswith(proxy):
                    doi = doi[len(proxy) :]
                    break
            if doi.count("/") == 0 or not doi.startswith("10."):
                print("    🤕 invalid DOI:", doi)
                valid = False
            entry["doi"] = doi
    return valid


def normalize_whitespace(entries: list[dict]):
    """Normalize the whitespace inside the field values."""
    for entry in entries:
        for key, value in list(entry.items()):
            entry[key] = re.sub(r"\s+", " ", value)


HYPFUN_REGEX = "[-֊־᠆‐‑‒––―⸺⸻﹘﹣－=᐀゠⸗⹀〜〰~⸚]"  # noqa: RUF001


def fix_page_double_hyphen(entries: list[dict]) -> bool:
    """Fix page ranges for which no double hyphen is used."""
    valid = True
    for entry in entries:
        pages = entry.get("pages")
        if pages is not None:
            parts = [part.strip() for part in re.split(HYPFUN_REGEX, pages)]
            parts = [part for part in parts if part != ""]
            if len(parts) == 1:
                entry["pages"] = parts[0]
            elif len(parts) == 2:
                entry["pages"] = "--".join(parts)
            else:
                print("    🤕 invalid page range:", pages)
                valid = False
    return valid


def abbreviate_journal_iso(entries: list[dict], custom: dict[str, str]):
    """Replace journal names by their ISO abbreviation."""
    abbreviator = Abbreviate.create()
    for entry in entries:
        journal = entry.get("journal")
        if journal is not None and "." not in journal:
            abbrev = custom.get(journal)
            if abbrev is None:
                abbrev = abbreviator(journal, remove_part=False)
            print(f"    📓 {journal} -> {abbrev}")
            abbrev = custom.get(abbrev, abbrev)
            entry["journal"] = abbrev


def merge_entries(entries: list[dict], field) -> bool:
    """Merge entries who have the same value for the given key."""
    lookup = {}
    missing_field = []
    merge_conflict = False
    for entry in entries:
        identifier = entry.get(field)
        if identifier is None:
            print(f"    👽 Cannot merge entry without {field}:", entry[KEY])
            missing_field.append(entry)
        else:
            other = lookup.setdefault(identifier)
            if other is None:
                lookup[identifier] = entry
            else:
                for key, value in entry.items():
                    if key not in other:
                        other[key] = value
                    elif key != KEY and other[key] != value:
                        print(f"    😭 Same {field}={identifier}, different {key}:")
                        print(f"        {value}")
                        print(f"        {other[key]}")
                        merge_conflict = True
                print(f"    🔗 Merged entries with same {field} = {identifier}")
    result = list(lookup.values()) + missing_field
    entries[:] = result
    return merge_conflict


def sort_entries(entries: list[dict]):
    """Sort the entries in convenient way: by year, then by author."""

    def keyfn(entry):
        return (
            entry.get("year", "0000")
            + entry.get("author", "Aaaa, Aaaa")
            + entry.get("title", "Title")
        )

    entries.sort(key=keyfn)


def write_output(entries: list[dict], fn_out: str, brace_title_words: bool, retcode: int) -> int:
    """Write out the fixed bibtex file, in case it has changed.

    Parameters
    ----------
    enties
        The list of BibTeX entries as dictionaries.
    fn_out
        The output filename.
    brace_title_words
        Whether to wrap words in titles that should preserve capitalization in braces.
    retcode
        The provisional return code, based on previous checks.

    Returns
    -------
    retcode
        The final return code.
    """
    if retcode == RETURN_CODE_CHANGED:
        # Write out a single BibTeX database.
        with tempfile.TemporaryDirectory("rr-bibsane") as dn_tmp:
            fn_tmp = os.path.join(dn_tmp, "tmp.bib")

            # Convert entry dictionaries back into pybtex entries.
            lib = pybtex.database.BibliographyData()
            for entry in entries:
                pyb_entry = pybtex.database.Entry(entry[ETYPE])
                fields = [
                    (key.lower(), value) for key, value in entry.items() if key not in (ETYPE, KEY)
                ]
                fields.sort()
                for key, value in fields:
                    encoded = U2L(value)
                    if brace_title_words and key in ("title", "journal"):
                        encoded = brace_words(encoded)
                    pyb_entry.fields[key] = encoded
                lib.add_entry(entry[KEY], pyb_entry)
            lib.to_file(fn_tmp)

            # Check if the file has changed.
            if os.path.isfile(fn_out):
                old_hash = compute_file_digest(fn_out)
                new_hash = compute_file_digest(fn_tmp)
                if old_hash == new_hash:
                    retcode = RETURN_CODE_SUCCESS
            if retcode == RETURN_CODE_CHANGED:
                print("💾 Please check the new or corrected file:", fn_out)
                shutil.copy(fn_tmp, fn_out)
            else:
                print("😀 No changes to", fn_out)
    else:
        print(f"💥 Broken bibliography. Not writing: {fn_out}")

    return retcode


NEEDS_BRACES = re.compile(r"^.+[A-Z].*$")


def brace_words(title: str) -> str:
    """Wrap words in braces if it seems useful."""
    result = []
    for word in title.split():
        if NEEDS_BRACES.match(word):
            result.append(f"{{{word}}}")
        else:
            result.append(word)
    return " ".join(result)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Actions perform the actual work in the worker processes of StepUp."""

import shlex

import attrs

from stepup.core.worker import WorkThread

from .add_notes_pdf import main as add_notes_pdf_main
from .bibsane import main as bibsane_main
from .cat_pdf import main as cat_pdf_main
from .check_hrefs import main as check_hrefs_main
from .compile_latex import main as compile_latex_main
from .compile_tectonic import main as compile_tectonic_main
from .compile_typst import main as compile_typst_main
from .convert_inkscape import main as convert_inkscape_main
from .convert_markdown import main as convert_markdown_main
from .convert_weasyprint import main as convert_weasyprint_main
from .execute_papermill import main as execute_papermill_main
from .flatten_latex import main as flatten_latex_main
from .make_inventory import main as make_inventory_main
from .nup_pdf import main as nup_pdf_main
from .raster_pdf import main as raster_pdf_main
from .sync_zenodo import main as sync_zenodo_main
from .unplot import main as unplot_main
from .wrap_git import main as wrap_git_main
from .zip_inventory import main as zip_inventory_main

__all__ = (
    "add_notes_pdf",
    "bibsane",
    "cat_pdf",
    "check_hrefs",
    "compile_latex",
    "compile_tectonic",
    "compile_typst",
    "convert_inkscape",
    "convert_markdown",
    "flatten_latex",
    "make_inventory",
    "nup_pdf",
    "raster_pdf",
    "sync_zenodo",
    "unplot",
    "wrap_git",
    "zip_inventory",
)


@attrs.define
class ActionWrapper:
    """Wrapper for main functions to get support the action API."""

    main: callable

    def __call__(self, argstr: str) -> int:
        self.main(shlex.split(argstr))
        return 0


add_notes_pdf = ActionWrapper(add_notes_pdf_main)
bibsane = ActionWrapper(bibsane_main)
cat_pdf = ActionWrapper(cat_pdf_main)
check_hrefs = ActionWrapper(check_hrefs_main)
convert_markdown = ActionWrapper(convert_markdown_main)
execute_papermill = ActionWrapper(execute_papermill_main)
flatten_latex = ActionWrapper(flatten_latex_main)
make_inventory = ActionWrapper(make_inventory_main)
nup_pdf = ActionWrapper(nup_pdf_main)
raster_pdf = ActionWrapper(raster_pdf_main)
sync_zenodo = ActionWrapper(sync_zenodo_main)
unplot = ActionWrapper(unplot_main)
zip_inventory = ActionWrapper(zip_inventory_main)


@attrs.define
class WorkThreadActionWrapper:
    """Wrapper for main functions to get support the action API."""

    main: callable

    def __call__(self, argstr: str, work_thread: WorkThread) -> int:
        self.main(shlex.split(argstr), work_thread)
        return 0


compile_latex = WorkThreadActionWrapper(compile_latex_main)
compile_tectonic = WorkThreadActionWrapper(compile_tectonic_main)
compile_typst = WorkThreadActionWrapper(compile_typst_main)
convert_inkscape = WorkThreadActionWrapper(convert_inkscape_main)
convert_weasyprint = WorkThreadActionWrapper(convert_weasyprint_main)
wrap_git = WorkThreadActionWrapper(wrap_git_main)



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Inventory utilities."""

import stat

import attrs
from path import Path

from stepup.core.hash import compute_file_digest


@attrs.define
class FileSummary:
    """Data structure for a record in an inventory text file.

    There are a few different ways in which files are summarized:

    - Directories are never included.
    - The properties of ordinary files are just computed.
      The digest is a 64-bit BLAKE2B hash.
    - Symbolic links are not followed.
      Their size is None and the digest is derived from the link destination path,
      not the file contents.
      Links to directories are supported.
    """

    size: int | None = attrs.field()
    mode: str = attrs.field()
    digest: bytes = attrs.field()
    path: str = attrs.field()


def get_summary(path: str, root: str) -> FileSummary:
    """Compute a file summary for an inventory file.

    Parameters
    ----------
    path
        The location of the file to be summarized,
        relative to the current working directory.
    root
        The parent of the inventory file, to construct relative paths.

    Returns
    -------
    file_summary
        Contains the size, mode, digest, and relative path of the file.
    """
    path = Path(path)
    st = path.stat(follow_symlinks=False)
    size = None if stat.S_ISLNK(st.st_mode) else st.st_size
    mode = stat.filemode(st.st_mode)
    digest = compute_file_digest(path, follow_symlinks=False)
    relpath = path.relpath(root).normpath()
    return FileSummary(size, mode, digest, relpath)


def format_summary(fs: FileSummary) -> str:
    """Create a single-line string representation of the summary."""
    size_str = "               " if fs.size is None else format(fs.size, "15d")
    return f"{size_str} {fs.mode} {fs.digest.hex()} {fs.path}"


def parse_summary(line: str) -> FileSummary:
    """Convert a single-line string back into a summary."""
    if len(line) < 157:
        raise ValueError(f"Line too short to be converted to a FileSummary: {line}.")
    size_str = line[:15]
    size = None if size_str == "               " else int(size_str)
    mode = line[16:26]
    digest = bytes.fromhex(line[27:155])
    path = line[156:].strip()
    return FileSummary(size, mode, digest, path)


def check_summary(new: FileSummary, ref: FileSummary):
    """Verify that the new digest is consistent with the reference."""
    if new.size != ref.size:
        raise ValueError(f"File size should be {ref.size} but got {new.size}: {new.path}")
    if new.mode != ref.mode:
        raise ValueError(f"File mode should be {ref.mode} but got {new.mode}: {new.path}")
    if new.digest != ref.digest:
        raise ValueError(f"File digest mismatch: {new.path}")
